divert(_NullStream)
dnl -----------------------------------------------
dnl   m4 file for bsd specific taxi features
dnl -----------------------------------------------

dnl note convention: initial lowercase after _ indicates function
dnl                  initial uppercase after _ indicates variable

dnl ###########################
dnl   list of global variables
dnl ###########################

define(`_Os', `bsd')
dnl _FopsPrototypes
dnl _FopsDefinitions
define(`_Success', `0')dnl
define(`_DevStruct', `_DriverName`'_dev')dnl

dnl ###########################
dnl   error codes
dnl ###########################
define(`_ENOMEM', `ENOMEM')

dnl ###########################
dnl   flags
dnl ###########################
define(`_WAITOK', `M_WAITOK')dnl tells malloc it is ok to wait for memory
define(`_NOWAIT', `M_NOWAIT')dnl tells malloc it is NOT ok to wait for memory

dnl ###########################
dnl   internal functions
dnl ###########################

dnl _addFopsPrototype(prototype)
dnl	adds the specified prototype to the list of prototypes in _FopsPrototypes
define(`_addFopsPrototype',dnl
`define(`_FopsPrototypes',dnl
defn(`_FopsPrototypes')
$1)')

dnl _addFile_ops_struct(method, methodName)
dnl     adds specified method and method name to file_operations data structure
define(`_addFile_ops_struct',dnl
`define(`_file_operations_list',dnl
defn(`_file_operations_list')dnl
`  $1 = $2,'
)')dnl

dnl _declareFopsPrototypes is a high level macro which spits out the prototypes for all
dnl	file operation functions which have been registered
define(`_declareFopsPrototypes', 
`/* Prototypes for registered file operation functions */
defn(`_FopsPrototypes')');

dnl _addFopsDefinition(definition)
dnl	adds the specified definition to the list of definitions in _FopsDefinitions
define(`_addFopsDefinition',dnl
`define(`_FopsDefinitions',dnl
defn(`_FopsDefinitions')
$1)')

dnl _ModuleEntryFunctionStream       Represents the stream which will buffer access point initialization
dnl                                  statements at the end of module source code on Bsd
dnl                                  e.g. void __init myModule_init(void) {...}
dnl                                       ...other code here...
dnl                                       module_init(myModule_init); //output this typically at end
dnl                                       EOF
define(`_ModuleEntryFunctionStream',`1')

dnl ########################################################################################
dnl _OsDefines defines all CPP symbols and vars that the target OS requires
dnl ########################################################################################
define(`_OsDefines',`')dnl

dnl _addOsDefine(define)
define(`_addOsDefine',dnl
`define(`_OsDefines',dnl
defn(`_OsDefines')
$1)')

dnl ########################################################################################
dnl _osIncludes causes all pertinent OS specific headers to be included
dnl 	CHANGE: at some point this macro needs to do something EXTREEMLY clever:
dnl 	it needs to read the input file and find out which files must be included
dnl	perhaps by detecting which symbols are used.  
dnl	this will probably be subtle, and quick to anger.
dnl ########################################################################################
define(`_osIncludes',`dnl
`#'include <sys/types.h>
`#'include <sys/module.h>
`#'include <sys/systm.h>  /* uprintf */
`#'include <sys/errno.h>
`#'include <sys/param.h>  /* defines used in kernel.h */
`#'include <sys/kernel.h> /* types used in module initialization */
`#'include <sys/conf.h>   /* cdevsw struct */
`#'include <sys/uio.h>    /* uio struct */
`#'include <sys/malloc.h>
');

dnl ########################################################################################
dnl _osVars causes all pertinent OS specific variables to be declared
dnl ########################################################################################
define(`_osVars',dnl
`/* OS specific vars */
static dev_t _DevStruct;
');

dnl ##########################################################################################
dnl _declareFopsStruct declares a file operations struct to be used in the driver's .c source
dnl 	the format of such a struct is obviously OS dependant.  under bsd it comes from fs.h
dnl ##########################################################################################
define(`_declareFopsStruct',`
/*************************************************************************
 * File operations structure declaration.
 * Function pointers are automatically set to names of functions as the
 * functions are registered in the m4 source.
 *************************************************************************/
static struct cdevsw _DriverName`'_fops = {
  .d_name = "_DriverName",
  .d_maj = _MajorNumber,
_file_operations_list};
')

dnl _defineFops is a high level macro which emits the definitions for all
dnl	file operation functions which have been registered
define(`_defineFops',
`/* Definitions for registered file operation functions */
defn(`_FopsDefinitions')');

dnl *********************************************************************
dnl *********************************************************************
dnl		    EXTERNALLY AVAILABLE FUNCTIONS
dnl *********************************************************************
dnl *********************************************************************

dnl ############################################################
dnl   init/exit function declarations
dnl   NOTE: intentional missing dnl's between divert statements
dnl ############################################################

dnl ##################################################
dnl   _load
dnl        > registers the function used to load
dnl          the module when requested by OS
dnl ##################################################
define(`_load',`dnl
define(`_LoadFN', `_DriverName`'_loader')dnl
define(`_LoadProto',dnl
`static int _LoadFN`'(struct module *m, int what, void *arg)')dnl
divert(_ModuleEntryFunctionStream)dnl
DEV_MODULE(_DriverName, _LoadFN, NULL);
divert(_StdoutStream)dnl
define(`_defineInitModule',`dnl
_LoadProto {
 int _MethodRetval = _Success;
 _DeclareTaxiRetval
 switch(what) {
 case MOD_`'LOAD:
$1
 break;
 case MOD_`'UNLOAD:
_ModUnload
 break;
 default:
    return EINVAL;
 }
 return _MethodRetval;
}
')dnl
')dnl

dnl ##################################################
dnl   _unload
dnl        > registers the function used to unload
dnl          the module when requested by OS
dnl ##################################################
define(`_unload',`dnl
define(`_ModUnload',`$1')dnl
')dnl
define(`_defineExitModule', `')dnl this is not needed in bsd, part of loader function

dnl #################################################
dnl FOR NOW, NOT EXPLICITLY NEEDED BECAUSE ON EOF, ALL STREAMS
dnl ARE OUTPUT TO STDOUT.  HOWEVER, LATER ON, IF WE USE 
dnl MULTIPLE DIVERSIONS, IT MAY BE USEFUL TO CONTROL THE
dnl ORDER IN WHICH THEY ARE OUTPUTTED
dnl
dnl undivert(_ModuleEntryFunctionStream)dnl
dnl #################################################

dnl _lseek(`codeBlock')  -registers a file operation function
define(`_lseek', `/*** I DON'T KNOW HOW TO LSEEK ***/')dnl

dnl _read(`codeBlock')  -registers a file operation function
define(`_read',dnl
`define(`_FopREADFN', `_DriverName`'_READ')dnl
define(`_FopREADProto',dnl
`static int _FopREADFN`'(dev_t dev, struct uio *uio, int ioflag)')dnl
_addFopsPrototype(_FopREADProto;)dnl
_addFopsDefinition(_FopREADProto {$1})dnl
_addFile_ops_struct(`.d_read', `_DriverName`'_READ')dnl
')

dnl _setWriteBytesWritten is a helper function for _write
define(`_setWriteBytesWritten', `define(`_WriteBytesWritten', `$1')')
dnl _write(`codeBlock')  -registers a file operation function
define(`_write',dnl
`define(`_FopWRITEFN', `_DriverName`'_WRITE')dnl
define(`_FopWRITEProto',dnl
`static int _FopWRITEFN`'(dev_t dev, struct uio *uio, int ioflag)')dnl # note, no semi-colon!
define(`_InputBase', `uio->uio_`'iov->iov_`'base')dnl
define(`_InputLength', `uio->uio_`'iov->iov_`'l`'en')dnl
_addFopsPrototype(_FopWRITEProto;)dnl
_addFopsDefinition(`_FopWRITEProto {
  _DeclareTaxiRetval
  int _MethodRetval = _Success;
$1
  _MethodRetval = _TaxiRetval;
  _return(_MethodRetval)
`}'')dnl
_addFile_ops_struct(`.d_write', `_DriverName`'_WRITE')dnl
')

dnl _open(`codeBlock')  -registers a file operation function
define(`_open',dnl
`define(`_FopOPENFN', `_DriverName`'_OPEN')dnl
define(`_FopOPENProto',dnl
`static int _FopOPENFN`'(dev_t dev, int oflags, int devtype, struct thread *p)')dnl 
_addFopsPrototype(_FopOPENProto;)dnl
_addFopsDefinition(_FopOPENProto {$1})dnl
_addFile_ops_struct(`.d_open', `_DriverName`'_OPEN')dnl
')

dnl _release(`codeBlock')  -registers a file operation function
define(`_release',dnl
`define(`_FopRELEASEFN', `_DriverName`'_RELEASE')dnl
define(`_FopRELEASEProto',dnl
`static int _FopRELEASEFN`'(dev_t dev, int fflag, int devtype, struct thread *p)')dnl
_addFopsPrototype(_FopRELEASEProto;)dnl
_addFopsDefinition(_FopRELEASEProto {$1})dnl
_addFile_ops_struct(`.d_close', `_DriverName`'_RELEASE')dnl
')

dnl _kernelPrint(`string')  -prints a string to the kernel logging facility
define(`_kernelPrint',`printf($1);')dnl

dnl _osLicense(`licenseType')  -emits OS specific license type macro
define(`_osLicense', `')dnl no licence registration under bsd, thank you.

dnl _setPortRange(`base',`range')  -allocates range of I/O starting at `base' for module
define(`_setPortRange',`//setPortRange definition')dnl

dnl _setMemoryRange(`base',`range')  -allocates range of I/O shared memory starting at `base' for module
define(`_setMemoryRange',`//setMemoryRange definition')dnl

dnl _unsetPortRange(`base',`range')  -allocates range of I/O starting at `base' for module
define(`_unsetPortRange', `//unsetPortRange definition')dnl

dnl _unsetMemoryRange(`base',`range')  -allocates range of I/O shared memory starting at `base' for module
define(`_unsetMemoryRange', `//unsetMemoryRange definition')dnl

dnl _markModuleUsage   -acquires semaphore preventing module removal
define(`_markModuleUsage', `')dnl bsd seems to do this automatically

dnl _unmarkModuleUsage   -releases semaphore allowing module removal
define(`_unmarkModuleUsage', `')dnl bsd seems to do this automatically

dnl _registerModule()  -registers device with the OS
dnl	returns: _Success
define(`_registerModule',dnl
`_DevStruct = make_dev(&`'_DriverName`'_fops, 
    _MinorNumber,
    UID_ROOT,
    GID_WHEEL,
    0666,
    "_DriverName");
  _TaxiRetval = _Success;')dnl

dnl _unregisterModule()  -unregisters device with the OS
define(`_unregisterModule', `destroy_dev(_DevStruct);')dnl

dnl _copyin(userland address, kernelspace address, count)
dnl    copies a string of length count from userland to kernelspace
define(`_copyin', `_TaxiRetval = copyin($1, $2, $3);')dnl

dnl _mallocDeclare(type)
dnl	declares a malloc type
define(`_mallocDeclare', `_addOsDefine(MALLOC_DECLARE($1);)')dnl

dnl _mallocDefine(type, shortdesc, longdesc)
dnl	defines a malloc type
define(`_mallocDefine',dnl
`_mallocDeclare($1)dnl
_addOsDefine(MALLOC_DEFINE($1, $2, $3);)')dnl

dnl _malloc(size, type, flags)
dnl	allocates memory of the appropriate size, and cast
define(`_malloc', `malloc($1, $2, $3)')dnl

dnl _free(addr, type)
dnl	frees memory allocated with _malloc
define(`_free', `free($1, $2)')dnl

divert(_StdoutStream)
